"Timeseries Module\n\nThis module mirrors the Timeseries API. It allows you to fetch data from the api and output it in various formats.\n\nhttps://doc.cognitedata.com/0.6/#Cognite-API-Time-series\n"
from cognite import _utils, config
from cognite.v05.dto import TimeSeriesResponse


def get_time_series_by_id(id, include_metadata=False, **kwargs):
    "Returns a TimeseriesResponse object containing the requested timeseries.\n\n    Args:\n        id (int):           ID of timeseries to look up\n\n        include_metadata (bool):    Decide if the metadata field should be returned or not. Defaults to False.\n\n    Keyword Arguments:\n        api_key (str):          Your api-key.\n\n        project (str):          Project name.\n\n    Returns:\n        v05.dto.TimeSeriesResponse: A data object containing the requested timeseries.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.6/projects/{}/timeseries/{}".format(project, id)
    headers = {"api-key": api_key, "accept": "application/json"}
    params = {"includeMetadata": include_metadata}
    res = _utils.get_request(url=url, headers=headers, params=params, cookies=config.get_cookies())
    return TimeSeriesResponse(res.json())


def get_multiple_time_series_by_id(ids, include_metadata=False, **kwargs):
    "Returns a TimeseriesResponse object containing the requested timeseries.\n\n    Args:\n        ids (List[int]):           IDs of timeseries to look up\n\n    Keyword Arguments:\n        api_key (str):          Your api-key.\n        project (str):          Project name.\n\n    Returns:\n        v05.dto.TimeSeriesResponse: A data object containing the requested timeseries with several getter methods with different\n        output formats.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.6/projects/{}/timeseries/byids".format(project)
    headers = {"api-key": api_key, "accept": "application/json", "content-type": "application/json"}
    body = {"items": ids}
    params = {"includeMetadata": include_metadata}
    res = _utils.post_request(url=url, body=body, params=params, headers=headers, cookies=config.get_cookies())
    return TimeSeriesResponse(res.json())


def search_for_time_series(
    name=None,
    description=None,
    query=None,
    unit=None,
    is_string=None,
    is_step=None,
    metadata=None,
    asset_ids=None,
    asset_subtrees=None,
    min_created_time=None,
    max_created_time=None,
    min_last_updated_time=None,
    max_last_updated_time=None,
    **kwargs
):
    'Returns a TimeSeriesResponse object containing the search results.\n\n    Args:\n        name (str): Prefix and fuzzy search on name.\n        description (str):  Prefix and fuzzy search on description.\n        query (str):    Search on name and description using wildcard search on each of the words (separated by spaces).\n                        Retrieves results where at least on word must match. Example: "some other"\n        unit (str): Filter on unit (case-sensitive)\n        is_string (bool): Filter on whether the ts is a string ts or not.\n        is_step (bool): Filter on whether the ts is a step ts or not.\n        metadata (Dict):    Filter out time series that do not match these metadata fields and values (case-sensitive).\n                            Format is {"key1": "val1", "key2", "val2"}\n        asset_ids (List): Filter out time series that are not linked to any of these assets. Format is [12,345,6,7890].\n        asset_subtrees (List):  Filter out time series that are not linked to assets in the subtree rooted at these assets.\n                                Format is [12,345,6,7890].\n        min_created_time (int):   Filter out time series with createdTime before this. Format is milliseconds since epoch.\n        max_created_time (int):   Filter out time series with createdTime after this. Format is milliseconds since epoch.\n        min_last_updated_time (int): Filter out time series with lastUpdatedTime before this. Format is milliseconds since epoch.\n        max_last_updated_time (int): Filter out time series with lastUpdatedTime after this. Format is milliseconds since epoch.\n\n    Keyword Arguments:\n        api_key (str):  Your api-key.\n        project (str):  Project name.\n        sort (str):     "createdTime" or "lastUpdatedTime". Field to be sorted.\n                        If not specified, results are sorted by relevance score.\n        dir (str):      "asc" or "desc". Only applicable if sort is specified. Default \'desc\'.\n        limit (int):    Return up to this many results. Maximum is 1000. Default is 25.\n        offset (int):   Offset from the first result. Sum of limit and offset must not exceed 1000. Default is 0.\n        boost_name (bool): Whether or not boosting name field. This option is experimental and can be changed.\n\n    Returns:\n        v05.dto.TimeSeriesResponse: A data object containing the requested timeseries with several getter methods with different\n        output formats.\n    '
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.6/projects/{}/timeseries/search".format(project)
    headers = {"api-key": api_key, "accept": "application/json", "content-type": "application/json"}
    params = {
        "name": name,
        "description": description,
        "query": query,
        "unit": unit,
        "isString": is_string,
        "isStep": is_step,
        "metadata": (str(metadata) if (metadata is not None) else None),
        "assetIds": (str(asset_ids) if (asset_ids is not None) else None),
        "assetSubtrees": (str(asset_subtrees) if (asset_subtrees is not None) else None),
        "minCreatedTime": min_created_time,
        "maxCreatedTime": max_created_time,
        "minLastUpdatedTime": min_last_updated_time,
        "maxLastUpdatedTime": max_last_updated_time,
        "sort": kwargs.get("sort"),
        "dir": kwargs.get("dir"),
        "limit": kwargs.get("limit"),
        "offset": kwargs.get("offset"),
        "boostName": kwargs.get("boost_name"),
    }
    res = _utils.get_request(url, params=params, headers=headers)
    return TimeSeriesResponse(res.json())



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api_timeseries_data_v2.proto

import sys

_b = sys.version_info[0] < 3 and (lambda x: x) or (lambda x: x.encode("latin1"))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="api_timeseries_data_v2.proto",
    package="api.v2",
    syntax="proto3",
    serialized_pb=_b(
        '\n\x1c\x61pi_timeseries_data_v2.proto\x12\x06\x61pi.v2"3\n\x0fStringDatapoint\x12\x11\n\ttimestamp\x18\x01 \x01(\x03\x12\r\n\x05value\x18\x02 \x01(\t"4\n\x10NumericDatapoint\x12\x11\n\ttimestamp\x18\x01 \x01(\x03\x12\r\n\x05value\x18\x02 \x01(\x01"?\n\x14StringTimeseriesData\x12\'\n\x06points\x18\x01 \x03(\x0b\x32\x17.api.v2.StringDatapoint"A\n\x15NumericTimeseriesData\x12(\n\x06points\x18\x01 \x03(\x0b\x32\x18.api.v2.NumericDatapoint"\x82\x01\n\x0eTimeseriesData\x12\x32\n\nstringData\x18\x01 \x01(\x0b\x32\x1c.api.v2.StringTimeseriesDataH\x00\x12\x34\n\x0bnumericData\x18\x02 \x01(\x0b\x32\x1d.api.v2.NumericTimeseriesDataH\x00\x42\x06\n\x04\x64\x61ta"\x95\x01\n\x13NamedTimeseriesData\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x32\n\nstringData\x18\x02 \x01(\x0b\x32\x1c.api.v2.StringTimeseriesDataH\x00\x12\x34\n\x0bnumericData\x18\x03 \x01(\x0b\x32\x1d.api.v2.NumericTimeseriesDataH\x00\x42\x06\n\x04\x64\x61ta"T\n\x18MultiNamedTimeseriesData\x12\x38\n\x13namedTimeseriesData\x18\x01 \x03(\x0b\x32\x1b.api.v2.NamedTimeseriesDataB1\n\x17\x63om.cognite.data.api.v2P\x01\xaa\x02\x13\x43ognite.Data.Api.V2b\x06proto3'
    ),
)


_STRINGDATAPOINT = _descriptor.Descriptor(
    name="StringDatapoint",
    full_name="api.v2.StringDatapoint",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="timestamp",
            full_name="api.v2.StringDatapoint.timestamp",
            index=0,
            number=1,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="value",
            full_name="api.v2.StringDatapoint.value",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=40,
    serialized_end=91,
)


_NUMERICDATAPOINT = _descriptor.Descriptor(
    name="NumericDatapoint",
    full_name="api.v2.NumericDatapoint",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="timestamp",
            full_name="api.v2.NumericDatapoint.timestamp",
            index=0,
            number=1,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="value",
            full_name="api.v2.NumericDatapoint.value",
            index=1,
            number=2,
            type=1,
            cpp_type=5,
            label=1,
            has_default_value=False,
            default_value=float(0),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=93,
    serialized_end=145,
)


_STRINGTIMESERIESDATA = _descriptor.Descriptor(
    name="StringTimeseriesData",
    full_name="api.v2.StringTimeseriesData",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="points",
            full_name="api.v2.StringTimeseriesData.points",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        )
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=147,
    serialized_end=210,
)


_NUMERICTIMESERIESDATA = _descriptor.Descriptor(
    name="NumericTimeseriesData",
    full_name="api.v2.NumericTimeseriesData",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="points",
            full_name="api.v2.NumericTimeseriesData.points",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        )
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=212,
    serialized_end=277,
)


_TIMESERIESDATA = _descriptor.Descriptor(
    name="TimeseriesData",
    full_name="api.v2.TimeseriesData",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="stringData",
            full_name="api.v2.TimeseriesData.stringData",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="numericData",
            full_name="api.v2.TimeseriesData.numericData",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="data", full_name="api.v2.TimeseriesData.data", index=0, containing_type=None, fields=[]
        )
    ],
    serialized_start=280,
    serialized_end=410,
)


_NAMEDTIMESERIESDATA = _descriptor.Descriptor(
    name="NamedTimeseriesData",
    full_name="api.v2.NamedTimeseriesData",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="name",
            full_name="api.v2.NamedTimeseriesData.name",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="stringData",
            full_name="api.v2.NamedTimeseriesData.stringData",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="numericData",
            full_name="api.v2.NamedTimeseriesData.numericData",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="data", full_name="api.v2.NamedTimeseriesData.data", index=0, containing_type=None, fields=[]
        )
    ],
    serialized_start=413,
    serialized_end=562,
)


_MULTINAMEDTIMESERIESDATA = _descriptor.Descriptor(
    name="MultiNamedTimeseriesData",
    full_name="api.v2.MultiNamedTimeseriesData",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="namedTimeseriesData",
            full_name="api.v2.MultiNamedTimeseriesData.namedTimeseriesData",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        )
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=564,
    serialized_end=648,
)

_STRINGTIMESERIESDATA.fields_by_name["points"].message_type = _STRINGDATAPOINT
_NUMERICTIMESERIESDATA.fields_by_name["points"].message_type = _NUMERICDATAPOINT
_TIMESERIESDATA.fields_by_name["stringData"].message_type = _STRINGTIMESERIESDATA
_TIMESERIESDATA.fields_by_name["numericData"].message_type = _NUMERICTIMESERIESDATA
_TIMESERIESDATA.oneofs_by_name["data"].fields.append(_TIMESERIESDATA.fields_by_name["stringData"])
_TIMESERIESDATA.fields_by_name["stringData"].containing_oneof = _TIMESERIESDATA.oneofs_by_name["data"]
_TIMESERIESDATA.oneofs_by_name["data"].fields.append(_TIMESERIESDATA.fields_by_name["numericData"])
_TIMESERIESDATA.fields_by_name["numericData"].containing_oneof = _TIMESERIESDATA.oneofs_by_name["data"]
_NAMEDTIMESERIESDATA.fields_by_name["stringData"].message_type = _STRINGTIMESERIESDATA
_NAMEDTIMESERIESDATA.fields_by_name["numericData"].message_type = _NUMERICTIMESERIESDATA
_NAMEDTIMESERIESDATA.oneofs_by_name["data"].fields.append(_NAMEDTIMESERIESDATA.fields_by_name["stringData"])
_NAMEDTIMESERIESDATA.fields_by_name["stringData"].containing_oneof = _NAMEDTIMESERIESDATA.oneofs_by_name["data"]
_NAMEDTIMESERIESDATA.oneofs_by_name["data"].fields.append(_NAMEDTIMESERIESDATA.fields_by_name["numericData"])
_NAMEDTIMESERIESDATA.fields_by_name["numericData"].containing_oneof = _NAMEDTIMESERIESDATA.oneofs_by_name["data"]
_MULTINAMEDTIMESERIESDATA.fields_by_name["namedTimeseriesData"].message_type = _NAMEDTIMESERIESDATA
DESCRIPTOR.message_types_by_name["StringDatapoint"] = _STRINGDATAPOINT
DESCRIPTOR.message_types_by_name["NumericDatapoint"] = _NUMERICDATAPOINT
DESCRIPTOR.message_types_by_name["StringTimeseriesData"] = _STRINGTIMESERIESDATA
DESCRIPTOR.message_types_by_name["NumericTimeseriesData"] = _NUMERICTIMESERIESDATA
DESCRIPTOR.message_types_by_name["TimeseriesData"] = _TIMESERIESDATA
DESCRIPTOR.message_types_by_name["NamedTimeseriesData"] = _NAMEDTIMESERIESDATA
DESCRIPTOR.message_types_by_name["MultiNamedTimeseriesData"] = _MULTINAMEDTIMESERIESDATA
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

StringDatapoint = _reflection.GeneratedProtocolMessageType(
    "StringDatapoint",
    (_message.Message,),
    dict(
        DESCRIPTOR=_STRINGDATAPOINT,
        __module__="api_timeseries_data_v2_pb2"
        # @@protoc_insertion_point(class_scope:api.v2.StringDatapoint)
    ),
)
_sym_db.RegisterMessage(StringDatapoint)

NumericDatapoint = _reflection.GeneratedProtocolMessageType(
    "NumericDatapoint",
    (_message.Message,),
    dict(
        DESCRIPTOR=_NUMERICDATAPOINT,
        __module__="api_timeseries_data_v2_pb2"
        # @@protoc_insertion_point(class_scope:api.v2.NumericDatapoint)
    ),
)
_sym_db.RegisterMessage(NumericDatapoint)

StringTimeseriesData = _reflection.GeneratedProtocolMessageType(
    "StringTimeseriesData",
    (_message.Message,),
    dict(
        DESCRIPTOR=_STRINGTIMESERIESDATA,
        __module__="api_timeseries_data_v2_pb2"
        # @@protoc_insertion_point(class_scope:api.v2.StringTimeseriesData)
    ),
)
_sym_db.RegisterMessage(StringTimeseriesData)

NumericTimeseriesData = _reflection.GeneratedProtocolMessageType(
    "NumericTimeseriesData",
    (_message.Message,),
    dict(
        DESCRIPTOR=_NUMERICTIMESERIESDATA,
        __module__="api_timeseries_data_v2_pb2"
        # @@protoc_insertion_point(class_scope:api.v2.NumericTimeseriesData)
    ),
)
_sym_db.RegisterMessage(NumericTimeseriesData)

TimeseriesData = _reflection.GeneratedProtocolMessageType(
    "TimeseriesData",
    (_message.Message,),
    dict(
        DESCRIPTOR=_TIMESERIESDATA,
        __module__="api_timeseries_data_v2_pb2"
        # @@protoc_insertion_point(class_scope:api.v2.TimeseriesData)
    ),
)
_sym_db.RegisterMessage(TimeseriesData)

NamedTimeseriesData = _reflection.GeneratedProtocolMessageType(
    "NamedTimeseriesData",
    (_message.Message,),
    dict(
        DESCRIPTOR=_NAMEDTIMESERIESDATA,
        __module__="api_timeseries_data_v2_pb2"
        # @@protoc_insertion_point(class_scope:api.v2.NamedTimeseriesData)
    ),
)
_sym_db.RegisterMessage(NamedTimeseriesData)

MultiNamedTimeseriesData = _reflection.GeneratedProtocolMessageType(
    "MultiNamedTimeseriesData",
    (_message.Message,),
    dict(
        DESCRIPTOR=_MULTINAMEDTIMESERIESDATA,
        __module__="api_timeseries_data_v2_pb2"
        # @@protoc_insertion_point(class_scope:api.v2.MultiNamedTimeseriesData)
    ),
)
_sym_db.RegisterMessage(MultiNamedTimeseriesData)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(
    descriptor_pb2.FileOptions(), _b("\n\027com.cognite.data.api.v2P\001\252\002\023Cognite.Data.Api.V2")
)
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
"""Models Module.

This module mirrors the Models API.

https://doc.cognitedata.com/0.6/models
"""
from typing import Any, Dict, List

import requests

from cognite import _utils as utils
from cognite import config


def create_model(
    name: str,
    description: str = "",
    metadata: Dict[str, Any] = None,
    input_fields: List[str] = None,
    output_fields: List[str] = None,
    **kwargs,
):
    """Creates a new model

    Args:
        name (str):             Name of model
        description (str):      Description
        metadata (Dict[str, Any]):          Metadata about model
        input_fields (List[str]):   List of input fields the model accepts
        output_fields (List[str]:   List of output fields the model produces

    Keyword Arguments:
        api_key (str):          Your api-key.
        project (str):          Project name.

    Returns:
        Dict: The created model.
    """
    api_key, project = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.6/projects/{}/analytics/models".format(project)
    headers = {"api-key": api_key, "accept": "application/json"}
    model_body = {
        "name": name,
        "description": description,
        "metadata": metadata or {},
        "inputFields": input_fields or [],
        "outputFields": output_fields or [],
    }
    res = utils.post_request(url, body=model_body, headers=headers, cookies=config.get_cookies())
    return res.json()


def get_models(**kwargs):
    """Get all models.

    Keyword Arguments:
        api_key (str):          Your api-key.
        project (str):          Project name.

    Returns:
        List[Dict]: List of models
    """
    api_key, project = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.6/projects/{}/analytics/models".format(project)
    headers = {"api-key": api_key, "accept": "application/json"}
    res = utils.get_request(url, headers=headers, cookies=config.get_cookies())
    return res.json()


def get_model(model_id: int, **kwargs):
    """Get a model by id.

    Args:
        model_id (int): Id of model to get.

    Keyword Arguments:
        api_key (str):          Your api-key.
        project (str):          Project name.

    Returns:
        Dict: The requested model
    """
    api_key, project = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.6/projects/{}/analytics/models/{}".format(project, model_id)
    headers = {"api-key": api_key, "accept": "application/json"}
    res = utils.get_request(url, headers=headers, cookies=config.get_cookies())
    return res.json()


def delete_model(model_id: int, **kwargs):
    """Delete a model.

    Will also delete all versions and schedules for this model.

    Args:
        model_id (int): Delete model with this id.

    Keyword Arguments:
        api_key (str):          Your api-key.
        project (str):          Project name.

    Returns:
        Dict: Empty Response
    """
    api_key, project = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.6/projects/{}/analytics/models/{}".format(project, model_id)
    headers = {"api-key": api_key, "accept": "application/json"}
    res = utils.delete_request(url, headers=headers, cookies=config.get_cookies())
    return res.json()


def train_model_version(
    model_id: int,
    name: str,
    source_package_id: int,
    train_source_package_id: int = None,
    metadata: Dict = None,
    description: str = None,
    args: Dict[str, Any] = None,
    scale_tier: str = None,
    machine_type: str = None,
    **kwargs,
):
    """Train a new version of a model.

    Args:
        model_id (int): Create a new version under the model with this id
        name (str): Name of model version. Must be unique on the model.
        source_package_id (int):    Use the source package with this id
        train_source_package_id (int):  Use this source package for training. If omitted, will default to
                                        source_package_id.
        metadata (Dict[str, Any]):  Metadata about model version
        description (str):  Description of model version
        args (Dict[str, Any]):   Dictionary of arguments to pass to the training job.
        scale_tier (str):   Which scale tier to use. Must be either "BASIC" or "CUSTOM"
        machine_type (str): Specify a machiene type Applies only if scale_tier is "CUSTOM".

    Keyword Arguments:
        api_key (str):          Your api-key.
        project (str):          Project name.

    Returns:
        Dict: The created model version.
    """
    api_key, project = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.6/projects/{}/analytics/models/{}/versions/train".format(project, model_id)
    body = {
        "name": name,
        "description": description or "",
        "sourcePackageId": source_package_id,
        "trainingDetails": {
            "sourcePackageId": train_source_package_id or source_package_id,
            "args": args or {},
            "scaleTier": scale_tier or "BASIC",
            "machineType": machine_type,
        },
        "metadata": metadata or {},
    }
    headers = {"api-key": api_key, "accept": "application/json"}
    res = utils.post_request(url, body=body, headers=headers, cookies=config.get_cookies())
    return res.json()


def get_versions(model_id: int, **kwargs):
    """Get all versions of a specific model.

    Args:
        model_id (int): Get versions for the model with this id.

    Keyword Arguments:
        api_key (str):          Your api-key.
        project (str):          Project name.

    Returns:
        List[Dict]: List of model versions
    """
    api_key, project = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.6/projects/{}/analytics/models/{}/versions".format(project, model_id)
    headers = {"api-key": api_key, "accept": "application/json"}
    res = utils.get_request(url, headers=headers, cookies=config.get_cookies())
    return res.json()


def get_version(model_id: int, version_id: int, **kwargs):
    """Get a specific model version by id.

    Args:
        model_id (int): Id of model which has the model version.
        version_id (int): Id of model version.

    Keyword Arguments:
        api_key (str):          Your api-key.
        project (str):          Project name.

    Returns:
        Dict: The requested model version
    """
    api_key, project = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.6/projects/{}/analytics/models/{}/versions/{}".format(
        project, model_id, version_id
    )
    headers = {"api-key": api_key, "accept": "application/json"}
    res = utils.get_request(url, headers=headers, cookies=config.get_cookies())
    return res.json()


def delete_version(model_id: int, version_id: int, **kwargs):
    """Delete a model version by id.

    Args:
        model_id (int): Id of model which has the model version.
        version_id (int): Id of model version.

    Keyword Arguments:
        api_key (str):          Your api-key.
        project (str):          Project name.

    Returns:
        Dict: The requested model version
    """
    api_key, project = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.6/projects/{}/analytics/models/{}/versions/{}".format(
        project, model_id, version_id
    )
    headers = {"api-key": api_key, "accept": "application/json"}
    res = utils.delete_request(url, headers=headers, cookies=config.get_cookies())
    return res.json()


def online_predict(
    model_id: int, version_id: int = None, instances: List = None, args: Dict[str, Any] = None, **kwargs
):
    """Perform online prediction on a models active version or a specified version.

    Args:
        model_id (int):     Perform a prediction on the model with this id. Will use active version.
        version_id (int):   Use this version instead of the active version. (optional)
        instances (List): List of JSON serializable instances to pass to your model one-by-one.
        args (Dict[str, Any])    Dictinoary of keyword arguments to pass to your predict method.

    Keyword Arguments:
        api_key (str):          Your api-key.
        project (str):          Project name.

    Returns:
        List: List of predictions for each instance.
    """
    api_key, project = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    if version_id:
        url = config.get_base_url() + "/api/0.6/projects/{}/analytics/models/{}/versions/{}/predict".format(
            project, model_id, version_id
        )
    else:
        url = config.get_base_url() + "/api/0.6/projects/{}/analytics/models/{}/predict".format(project, model_id)

    body = {"instances": instances, "args": args or {}}
    headers = {"api-key": api_key, "accept": "application/json"}
    res = utils.put_request(url, body=body, headers=headers, cookies=config.get_cookies())
    return res.json()


def create_source_package(
    name: str,
    package_name: str,
    available_operations: List[str],
    runtime_version: str,
    description: str = None,
    meta_data: Dict = None,
    file_path: str = None,
    **kwargs,
):
    """Upload a source package to the model hosting environment.

    Args:
        name (str): Name of source package
        package_name (str): name of root package for model
        available_operations (List[str]): List of routines which this source package supports ["predict", "train"]
        runtime_version (str): Version of environment in which the source-package should run. Currently only 0.1.
        description (str): Description for source package
        meta_data (Dict): User defined key value pair of additional information.
        file_path (str): File path of source package distribution. If not specified, a download url will be returned.

    Keyword Arguments:
        api_key (str):          Your api-key.
        project (str):          Project name.

    Returns:
        Dict: Source package ID if file path was specified. Else, source package id and upload url.

    """
    api_key, project = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.6/projects/{}/analytics/models/sourcepackages".format(project)
    body = {
        "name": name,
        "description": description or "",
        "packageName": package_name,
        "availableOperations": available_operations,
        "metadata": meta_data or {},
        "runtimeVersion": runtime_version,
    }
    headers = {"api-key": api_key, "accept": "application/json"}
    res = utils.post_request(url, body=body, headers=headers, cookies=config.get_cookies())
    if file_path:
        _upload_file(res.json().get("uploadUrl"), file_path)
        del res.json()["uploadUrl"]
        return res.json()
    return res.json()


def _upload_file(upload_url, file_path):
    with open(file_path, "rb") as fh:
        mydata = fh.read()
        response = requests.put(upload_url, data=mydata)
    return response


def get_source_packages(**kwargs):
    """Get all model source packages.

    Keyword Arguments:
        api_key (str):          Your api-key.
        project (str):          Project name.

    Returns:
        List[Dict]: List of source packages.
    """
    api_key, project = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.6/projects/{}/analytics/models/sourcepackages".format(project)
    headers = {"api-key": api_key, "accept": "application/json"}
    res = utils.get_request(url, headers=headers, cookies=config.get_cookies())
    return res.json()


def get_source_package(source_package_id: int, **kwargs):
    """Get model source package by id.

    Args:
        source_package_id (int): Id of soure package to get.

    Keyword Arguments:
        api_key (str):          Your api-key.
        project (str):          Project name.

    Returns:
        Dict: The requested source package.
    """
    api_key, project = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.6/projects/{}/analytics/models/sourcepackages/{}".format(
        project, source_package_id
    )
    headers = {"api-key": api_key, "accept": "application/json"}
    res = utils.get_request(url, headers=headers, cookies=config.get_cookies())
    return res.json()


def delete_source_package(source_package_id: int, **kwargs):
    """Delete source package by id.

    Args:
        source_package_id (int): Id of soure package to delete.

    Keyword Arguments:
        api_key (str):          Your api-key.
        project (str):          Project name.

    Returns:
        Dict: Empty response.
    """
    api_key, project = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.6/projects/{}/analytics/models/sourcepackages/{}".format(
        project, source_package_id
    )
    headers = {"api-key": api_key, "accept": "application/json"}
    res = utils.delete_request(url, headers=headers, cookies=config.get_cookies())
    return res.json()


def create_schedule(
    model_id: int,
    name: str,
    output_data_spec: Dict,
    input_data_spec: Dict,
    description: str = None,
    args: Dict = None,
    metadata: Dict = None,
    **kwargs,
):
    """Create a new schedule on a given model.

    Args:
        model_id (int): Id of model to create schedule on
        name (str): Name of schedule
        output_data_spec (Dict): Specification of output. Example below.
        input_data_spec (Dict): Specification of input. Example below.
        description (str): Description for schedule
        args (Dict): Dictionary of keyword arguments to pass to predict method.
        metadata (Dict): Dictionary of metadata about schedule

    Keyword Arguments:
        api_key (str):          Your api-key.
        project (str):          Project name.

    Returns:
        Dict: The created schedule.

    Examples
        The output data spec must look like this::

            {
                "timeSeries": [
                    {
                        "label": "string",
                        "id": 123456789
                    }
                ]
            }

        The input data spec must look like this. The local aggregate and the missingDataStrategy fields are optional::

            {
                "windowSize": "1s",
                "stride": "1s",
                "missingDataStrategy": "",
                "timeSeries": [
                  {
                    "label": "string",
                    "id": 0,
                    "missingDataStrategy": "string",
                    "aggregate": "string"
                  }
                ],
                "aggregate": "string",
                "granularity": "string"
            }
    """
    api_key, project = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.6/projects/{}/analytics/models/schedules".format(project)
    body = {
        "name": name,
        "description": description,
        "modelId": model_id,
        "args": args or {},
        "inputDataSpec": input_data_spec,
        "outputDataSpec": output_data_spec,
        "metadata": metadata or {},
    }
    headers = {"api-key": api_key, "accept": "application/json"}
    res = utils.post_request(url, body=body, headers=headers, cookies=config.get_cookies())
    return res.json()


def delete_schedule(schedule_id: int, **kwargs):
    """Delete a schedule by id.

    Args:
        schedule_id (int):  The id of the schedule to delete.

    Keyword Arguments:
        api_key (str):          Your api-key.
        project (str):          Project name.

    Returns:
        Dict: Empty response
    """
    api_key, project = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.6/projects/{}/analytics/models/schedules/{}".format(project, schedule_id)
    headers = {"api-key": api_key, "accept": "application/json"}
    res = utils.delete_request(url=url, headers=headers)
    return res.json()


def get_schedules(**kwargs):
    """Get all schedules.

    Keyword Arguments:
        api_key (str):          Your api-key.
        project (str):          Project name.

    Returns:
        List[Dict]: The requested schedules.
    """
    api_key, project = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.6/projects/{}/analytics/models/schedules".format(project)
    headers = {"api-key": api_key, "accept": "application/json"}
    res = utils.get_request(url=url, headers=headers)
    return res.json()


def get_schedule(schedule_id: int, **kwargs):
    """Get a schedule by id.

    Keyword Arguments:
        api_key (str):          Your api-key.
        project (str):          Project name.

    Returns:
        Dict: The requested schedule.
    """
    api_key, project = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.6/projects/{}/analytics/models/schedules/{}".format(project, schedule_id)
    headers = {"api-key": api_key, "accept": "application/json"}
    res = utils.get_request(url=url, headers=headers)
    return res.json()



"Tag Matching Module\n\nThis module mirrors the Tag Matching API. It allows the user to search for tag id matches.\n\nhttps://doc.cognitedata.com/0.4/#Cognite-API-Tag-Matching\n"
import cognite._utils as _utils
import cognite.config as config
from cognite.v04.dto import TagMatchingResponse


def tag_matching(tag_ids, fuzzy_threshold=0, platform=None, **kwargs):
    "Returns a TagMatchingObject containing a list of matched tags for the given query.\n\n    This method takes an arbitrary string as argument and performs fuzzy matching with a user defined threshold\n    toward tag ids in the system.\n\n    Args:\n        tag_ids (list):         The tag_ids to retrieve matches for.\n\n        fuzzy_threshold (int):  The threshold to use when searching for matches. A fuzzy threshold of 0 means you only\n                                want to accept perfect matches. Must be >= 0.\n\n        platform (str):         The platform to search on.\n\n    Keyword Args:\n        api_key (str):          Your api-key.\n\n        project (str):          Project name.\n\n    Returns:\n        v04.dto.TagMatchingResponse: A data object containing the requested data with several getter methods with different\n        output formats.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.4/projects/{}/tagmatching".format(project)
    body = {"tagIds": tag_ids, "metadata": {"fuzzyThreshold": fuzzy_threshold, "platform": platform}}
    headers = {"api-key": api_key, "content-type": "*/*", "accept": "application/json"}
    res = _utils.post_request(url=url, body=body, headers=headers, cookies=config.get_cookies())
    return TagMatchingResponse(res.json())



"Timeseries Module\n\nThis module mirrors the Timeseries API. It allows you to fetch data from the api and output it in various formats.\n\nhttps://doc.cognitedata.com/0.5/#Cognite-API-Time-series\n"
import io
import time
from concurrent.futures import ThreadPoolExecutor as Pool
from functools import partial
from urllib.parse import quote
import pandas as pd
import cognite._constants as _constants
import cognite._utils as _utils
import cognite.config as config
from cognite.auxiliary._protobuf_descriptors import _api_timeseries_data_v2_pb2
from cognite.v05.dto import (
    Datapoint,
    DatapointsResponse,
    DatapointsResponseIterator,
    LatestDatapointResponse,
    TimeSeries,
    TimeSeriesResponse,
    TimeseriesWithDatapoints,
)


def get_datapoints(name, start, end=None, aggregates=None, granularity=None, **kwargs):
    "Returns a DatapointsObject containing a list of datapoints for the given query.\n\n    This method will automate paging for the user and return all data for the given time period.\n\n    Args:\n        name (str):             The name of the timeseries to retrieve data for.\n\n        start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.\n                                    E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since\n                                    epoch or a datetime object which will be converted to ms since epoch UTC.\n\n        end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.\n\n        aggregates (list):      The list of aggregate functions you wish to apply to the data. Valid aggregate functions\n                                are: 'average/avg, max, min, count, sum, interpolation/int, stepinterpolation/step'.\n\n        granularity (str):      The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,\n                                second/s', or a multiple of these indicated by a number as a prefix e.g. '12hour'.\n\n    Keyword Arguments:\n        include_outside_points (bool):      No description\n\n        protobuf (bool):        Download the data using the binary protobuf format. Only applicable when getting raw data.\n                                Defaults to True.\n\n        processes (int):        Number of download processes to run in parallell. Defaults to number returned by cpu_count().\n\n        api_key (str):          Your api-key.\n\n        project (str):          Project name.\n\n        limit (str):            Max number of datapoints to return. If limit is specified, this method will not automate\n                                paging and will return a maximum of 100,000 dps.\n\n    Returns:\n        v05.dto.DatapointsResponse: A data object containing the requested data with several getter methods with different\n        output formats.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    (start, end) = _utils.interval_to_ms(start, end)
    if kwargs.get("limit"):
        return _get_datapoints_user_defined_limit(
            name,
            aggregates,
            granularity,
            start,
            end,
            limit=kwargs.get("limit"),
            protobuf=kwargs.get("protobuf"),
            include_outside_points=kwargs.get("include_outside_points", False),
            api_key=api_key,
            project=project,
        )
    diff = end - start
    num_of_processes = kwargs.get("processes", _constants.NUM_OF_WORKERS)
    granularity_ms = 1
    if granularity:
        granularity_ms = _utils.granularity_to_ms(granularity)
    steps = min(num_of_processes, max(1, int((diff / granularity_ms))))
    step_size = _utils.round_to_nearest(int((diff / steps)), base=granularity_ms)
    step_starts = [(start + (i * step_size)) for i in range(steps)]
    args = [{"start": start, "end": (start + step_size)} for start in step_starts]
    partial_get_dps = partial(
        _get_datapoints_helper_wrapper,
        name=name,
        aggregates=aggregates,
        granularity=granularity,
        protobuf=kwargs.get("protobuf", True),
        include_outside_points=kwargs.get("include_outside_points", False),
        api_key=api_key,
        project=project,
    )
    with Pool(steps) as p:
        datapoints = p.map(partial_get_dps, args)
    concat_dps = []
    [concat_dps.extend(el) for el in datapoints]
    return DatapointsResponse({"data": {"items": [{"name": name, "datapoints": concat_dps}]}})


def _get_datapoints_helper_wrapper(
    args, name, aggregates, granularity, protobuf, include_outside_points, api_key, project
):
    return _get_datapoints_helper(
        name,
        aggregates,
        granularity,
        args["start"],
        args["end"],
        protobuf=protobuf,
        api_key=api_key,
        project=project,
        include_outside_points=include_outside_points,
    )


def _get_datapoints_helper(name, aggregates=None, granularity=None, start=None, end=None, **kwargs):
    "Returns a list of datapoints for the given query.\n\n    This method will automate paging for the given time period.\n\n    Args:\n        name (str):       The name of the timeseries to retrieve data for.\n\n        aggregates (list):      The list of aggregate functions you wish to apply to the data. Valid aggregate functions\n                                are: 'average/avg, max, min, count, sum, interpolation/int, stepinterpolation/step'.\n\n        granularity (str):      The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,\n                                second/s', or a multiple of these indicated by a number as a prefix e.g. '12hour'.\n\n        start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.\n                                    E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since\n                                    epoch or a datetime object which will be converted to ms since epoch UTC.\n\n        end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.\n\n    Keyword Arguments:\n        include_outside_points (bool):  No description.\n\n        protobuf (bool):        Download the data using the binary protobuf format. Only applicable when getting raw data.\n                                Defaults to True.\n\n        api_key (str):          Your api-key. Obligatory in this helper method.\n\n        project (str):          Project name. Obligatory in this helper method.\n\n    Returns:\n        list of datapoints: A list containing datapoint dicts.\n    "
    (api_key, project) = (kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.5/projects/{}/timeseries/data/{}".format(project, quote(name, safe=""))
    use_protobuf = kwargs.get("protobuf", True) and (aggregates is None)
    limit = _constants.LIMIT if (aggregates is None) else _constants.LIMIT_AGG
    params = {
        "aggregates": aggregates,
        "granularity": granularity,
        "limit": limit,
        "start": start,
        "end": end,
        "includeOutsidePoints": kwargs.get("include_outside_points", False),
    }
    headers = {"api-key": api_key, "accept": ("application/protobuf" if use_protobuf else "application/json")}
    datapoints = []
    while ((not datapoints) or (len(datapoints[(-1)]) == limit)) and (params["end"] > params["start"]):
        res = _utils.get_request(url, params=params, headers=headers)
        if use_protobuf:
            ts_data = _api_timeseries_data_v2_pb2.TimeseriesData()
            ts_data.ParseFromString(res.content)
            res = [{"timestamp": p.timestamp, "value": p.value} for p in ts_data.numericData.points]
        else:
            res = res.json()["data"]["items"][0]["datapoints"]
        if not res:
            break
        datapoints.append(res)
        latest_timestamp = int(datapoints[(-1)][(-1)]["timestamp"])
        params["start"] = latest_timestamp + (_utils.granularity_to_ms(granularity) if granularity else 1)
    dps = []
    [dps.extend(el) for el in datapoints]
    return dps


def _get_datapoints_user_defined_limit(name, aggregates, granularity, start, end, limit, **kwargs):
    "Returns a DatapointsResponse object with the requested data.\n\n    No paging or parallelizing is done.\n\n    Args:\n        name (str):       The name of the timeseries to retrieve data for.\n\n        aggregates (list):      The list of aggregate functions you wish to apply to the data. Valid aggregate functions\n                                are: 'average/avg, max, min, count, sum, interpolation/int, stepinterpolation/step'.\n\n        granularity (str):      The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,\n                                second/s', or a multiple of these indicated by a number as a prefix e.g. '12hour'.\n\n        start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.\n                                    E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since\n                                    epoch or a datetime object which will be converted to ms since epoch UTC.\n\n        end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.\n\n        limit (str):            Max number of datapoints to return. Max is 100,000.\n\n    Keyword Arguments:\n        include_outside_points (bool):  No description.\n\n        protobuf (bool):        Download the data using the binary protobuf format. Only applicable when getting raw data.\n                                Defaults to True.\n\n        api_key (str):          Your api-key. Obligatory in this helper method.\n\n        project (str):          Project name. Obligatory in this helper method.\n    Returns:\n        v05.dto.DatapointsResponse: A data object containing the requested data with several getter methods with different\n        output formats.\n    "
    (api_key, project) = (kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.5/projects/{}/timeseries/data/{}".format(project, quote(name, safe=""))
    use_protobuf = kwargs.get("protobuf", True) and (aggregates is None)
    params = {
        "aggregates": aggregates,
        "granularity": granularity,
        "limit": limit,
        "start": start,
        "end": end,
        "includeOutsidePoints": kwargs.get("include_outside_points", False),
    }
    headers = {"api-key": api_key, "accept": ("application/protobuf" if use_protobuf else "application/json")}
    res = _utils.get_request(url, params=params, headers=headers)
    if use_protobuf:
        ts_data = _api_timeseries_data_v2_pb2.TimeseriesData()
        ts_data.ParseFromString(res.content)
        res = [{"timestamp": p.timestamp, "value": p.value} for p in ts_data.numericData.points]
    else:
        res = res.json()["data"]["items"][0]["datapoints"]
    return DatapointsResponse({"data": {"items": [{"name": name, "datapoints": res}]}})


def _split_TimeseriesWithDatapoints_if_over_limit(timeseries_with_datapoints, limit):
    "Takes a TimeseriesWithDatapoints and splits it into multiple so that each has a max number of datapoints equal\n    to the limit given.\n\n    Args:\n        timeseries_with_datapoints (v05.dto.TimeseriesWithDatapoints): The timeseries with data to potentially split up.\n\n    Returns:\n        A list of v05.dto.TimeSeriesWithDatapoints where each has a maximum number of datapoints equal to the limit given.\n    "
    timeseries_with_datapoints_list = []
    if len(timeseries_with_datapoints.datapoints) > limit:
        i = 0
        while i < len(timeseries_with_datapoints.datapoints):
            timeseries_with_datapoints_list.append(
                TimeseriesWithDatapoints(
                    name=timeseries_with_datapoints.name,
                    datapoints=timeseries_with_datapoints.datapoints[i : (i + limit)],
                )
            )
            i += limit
    else:
        timeseries_with_datapoints_list.append(timeseries_with_datapoints)
    return timeseries_with_datapoints_list


def post_multi_tag_datapoints(timeseries_with_datapoints, **kwargs):
    "Insert data into multiple timeseries.\n\n    Args:\n        timeseries_with_datapoints (List[v05.dto.TimeseriesWithDatapoints]): The timeseries with data to insert.\n\n    Keyword Args:\n        api_key (str): Your api-key.\n\n        project (str): Project name.\n\n        use_gzip (bool): Whether or not to gzip the request\n\n    Returns:\n        An empty response.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.5/projects/{}/timeseries/data".format(project)
    use_gzip = kwargs.get("use_gzip", False)
    headers = {"api-key": api_key, "content-type": "application/json", "accept": "application/json"}
    ul_dps_limit = 100000
    timeseries_with_datapoints_limited = []
    for entry in timeseries_with_datapoints:
        timeseries_with_datapoints_limited.extend(_split_TimeseriesWithDatapoints_if_over_limit(entry, ul_dps_limit))
    timeseries_to_upload_binned = _utils.first_fit(
        list_items=timeseries_with_datapoints_limited, max_size=ul_dps_limit, get_count=(lambda x: len(x.datapoints))
    )
    for bin in timeseries_to_upload_binned:
        body = {
            "items": [
                {"name": ts_with_data.name, "datapoints": [dp.__dict__ for dp in ts_with_data.datapoints]}
                for ts_with_data in bin
            ]
        }
        res = _utils.post_request(url, body=body, headers=headers, use_gzip=use_gzip)
    return res.json()


def post_datapoints(name, datapoints, **kwargs):
    "Insert a list of datapoints.\n\n    Args:\n        name (str):       Name of timeseries to insert to.\n\n        datapoints (list[v05.dto.Datapoint): List of datapoint data transfer objects to insert.\n\n    Keyword Args:\n        api_key (str): Your api-key.\n\n        project (str): Project name.\n\n    Returns:\n        An empty response.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.5/projects/{}/timeseries/data/{}".format(project, quote(name, safe=""))
    headers = {"api-key": api_key, "content-type": "application/json", "accept": "application/json"}
    ul_dps_limit = 100000
    i = 0
    while i < len(datapoints):
        body = {"items": [dp.__dict__ for dp in datapoints[i : (i + ul_dps_limit)]]}
        res = _utils.post_request(url, body=body, headers=headers)
        i += ul_dps_limit
    return res.json()


def get_latest(name, **kwargs):
    "Returns a LatestDatapointObject containing the latest datapoint for the given timeseries.\n\n    Args:\n        name (str):       The name of the timeseries to retrieve data for.\n\n    Keyword Arguments:\n        api_key (str):          Your api-key.\n\n        project (str):          Project name.\n\n    Returns:\n        v05.dto.LatestDatapointsResponse: A data object containing the requested data with several getter methods with different\n        output formats.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.5/projects/{}/timeseries/latest/{}".format(project, quote(name, safe=""))
    headers = {"api-key": api_key, "accept": "application/json"}
    res = _utils.get_request(url, headers=headers, cookies=config.get_cookies())
    return LatestDatapointResponse(res.json())


def get_multi_time_series_datapoints(datapoints_queries, start, end=None, aggregates=None, granularity=None, **kwargs):
    "Returns a list of DatapointsObjects each of which contains a list of datapoints for the given timeseries.\n\n    This method will automate paging for the user and return all data for the given time period(s).\n\n    Args:\n        datapoints_queries (list[v05.dto.DatapointsQuery]): The list of DatapointsQuery objects specifying which\n                                                                    timeseries to retrieve data for.\n\n        start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.\n                                    E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since\n                                    epoch or a datetime object which will be converted to ms since epoch UTC.\n\n        end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.\n\n        aggregates (list, optional):    The list of aggregate functions you wish to apply to the data. Valid aggregate\n                                        functions are: 'average/avg, max, min, count, sum, interpolation/int,\n                                        stepinterpolation/step'.\n\n        granularity (str):              The granularity of the aggregate values. Valid entries are : 'day/d, hour/h,\n                                        minute/m, second/s', or a multiple of these indicated by a number as a prefix\n                                        e.g. '12hour'.\n\n    Keyword Arguments:\n        include_outside_points (bool):  No description.\n\n        api_key (str):                  Your api-key.\n\n        project (str):                  Project name.\n\n    Returns:\n        list(v05.dto.DatapointsResponse): A list of data objects containing the requested data with several getter methods\n        with different output formats.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.5/projects/{}/timeseries/dataquery".format(project)
    (start, end) = _utils.interval_to_ms(start, end)
    num_of_dpqs_with_agg = 0
    num_of_dpqs_raw = 0
    for dpq in datapoints_queries:
        if ((dpq.aggregates is None) and (aggregates is None)) or (dpq.aggregates == ""):
            num_of_dpqs_raw += 1
        else:
            num_of_dpqs_with_agg += 1
    items = []
    for dpq in datapoints_queries:
        if (dpq.aggregates is None) and (aggregates is None):
            dpq.limit = int((_constants.LIMIT / num_of_dpqs_raw))
        else:
            dpq.limit = int((_constants.LIMIT_AGG / num_of_dpqs_with_agg))
        items.append(dpq.__dict__)
    body = {
        "items": items,
        "aggregates": (",".join(aggregates) if (aggregates is not None) else None),
        "granularity": granularity,
        "start": start,
        "includeOutsidePoints": kwargs.get("include_outside_points", False),
        "end": end,
    }
    headers = {"api-key": api_key, "content-type": "application/json", "accept": "application/json"}
    datapoints_responses = []
    has_incomplete_requests = True
    while has_incomplete_requests:
        res = _utils.post_request(url=url, body=body, headers=headers, cookies=config.get_cookies()).json()["data"][
            "items"
        ]
        datapoints_responses.append(res)
        has_incomplete_requests = False
        for (i, dpr) in enumerate(res):
            dpq = datapoints_queries[i]
            if len(dpr["datapoints"]) == dpq.limit:
                has_incomplete_requests = True
                latest_timestamp = dpr["datapoints"][(-1)]["timestamp"]
                ts_granularity = granularity if (dpq.granularity is None) else dpq.granularity
                next_start = latest_timestamp + (_utils.granularity_to_ms(ts_granularity) if ts_granularity else 1)
            else:
                next_start = end - 1
                if datapoints_queries[i].end:
                    next_start = datapoints_queries[i].end - 1
            datapoints_queries[i].start = next_start
    results = [{"data": {"items": [{"name": dpq.name, "datapoints": []}]}} for dpq in datapoints_queries]
    for res in datapoints_responses:
        for (i, ts) in enumerate(res):
            results[i]["data"]["items"][0]["datapoints"].extend(ts["datapoints"])
    return DatapointsResponseIterator([DatapointsResponse(result) for result in results])


def get_datapoints_frame(time_series, aggregates, granularity, start, end=None, **kwargs):
    "Returns a pandas dataframe of datapoints for the given timeseries all on the same timestamps.\n\n    This method will automate paging for the user and return all data for the given time period.\n\n    Args:\n        time_series (list):  The list of timeseries names to retrieve data for. Each timeseries can be either a string\n                            containing the timeseries or a dictionary containing the names of thetimeseries and a\n                            list of specific aggregate functions.\n\n        aggregates (list):  The list of aggregate functions you wish to apply to the data for which you have not\n                            specified an aggregate function. Valid aggregate functions are: 'average/avg, max, min,\n                            count, sum, interpolation/int, stepinterpolation/step'.\n\n        granularity (str):  The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,\n                            second/s', or a multiple of these indicated by a number as a prefix e.g. '12hour'.\n\n        start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.\n                                    E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since\n                                    epoch or a datetime object which will be converted to ms since epoch UTC.\n\n        end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.\n\n    Keyword Arguments:\n        api_key (str): Your api-key.\n\n        project (str): Project name.\n\n        cookies (dict): Cookies.\n\n        limit (str): Max number of rows to return. If limit is specified, this method will not automate\n                        paging and will return a maximum of 100,000 rows.\n        \n        processes (int):    Number of download processes to run in parallell. Defaults to number returned by cpu_count().\n\n    Returns:\n        pandas.DataFrame: A pandas dataframe containing the datapoints for the given timeseries. The datapoints for all the\n        timeseries will all be on the same timestamps.\n\n    Examples:\n        The ``timeseries`` parameter can take a list of strings and/or dicts on the following formats::\n\n            Using strings:\n                ['<timeseries1>', '<timeseries2>']\n\n            Using dicts:\n                [{'name': '<timeseries1>', 'aggregates': ['<aggfunc1>', '<aggfunc2>']},\n                {'name': '<timeseries2>', 'aggregates': []}]\n\n            Using both:\n                ['<timeseries1>', {'name': '<timeseries2>', 'aggregates': ['<aggfunc1>', '<aggfunc2>']}]\n    "
    if not isinstance(time_series, list):
        raise _utils.InputError("time_series should be a list")
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    cookies = config.get_cookies(kwargs.get("cookies"))
    (start, end) = _utils.interval_to_ms(start, end)
    if kwargs.get("limit"):
        return _get_datapoints_frame_user_defined_limit(
            time_series,
            aggregates,
            granularity,
            start,
            end,
            limit=kwargs.get("limit"),
            api_key=api_key,
            project=project,
            cookies=cookies,
        )
    diff = end - start
    num_of_processes = kwargs.get("processes") or _constants.NUM_OF_WORKERS
    granularity_ms = 1
    if granularity:
        granularity_ms = _utils.granularity_to_ms(granularity)
    steps = min(num_of_processes, max(1, int((diff / granularity_ms))))
    step_size = _utils.round_to_nearest(int((diff / steps)), base=granularity_ms)
    step_starts = [(start + (i * step_size)) for i in range(steps)]
    args = [{"start": start, "end": (start + step_size)} for start in step_starts]
    partial_get_dpsf = partial(
        _get_datapoints_frame_helper_wrapper,
        time_series=time_series,
        aggregates=aggregates,
        granularity=granularity,
        api_key=api_key,
        project=project,
        cookies=cookies,
    )
    if steps == 1:
        return _get_datapoints_frame_helper(
            time_series, aggregates, granularity, start, end, api_key=api_key, project=project, cookies=cookies
        )
    with Pool(steps) as p:
        dataframes = p.map(partial_get_dpsf, args)
    df = pd.concat(dataframes).drop_duplicates(subset="timestamp").reset_index(drop=True)
    return df


def _get_datapoints_frame_helper_wrapper(args, time_series, aggregates, granularity, api_key, project, cookies):
    return _get_datapoints_frame_helper(
        time_series,
        aggregates,
        granularity,
        args["start"],
        args["end"],
        api_key=api_key,
        project=project,
        cookies=cookies,
    )


def _get_datapoints_frame_helper(time_series, aggregates, granularity, start=None, end=None, **kwargs):
    "Returns a pandas dataframe of datapoints for the given timeseries all on the same timestamps.\n\n    This method will automate paging for the user and return all data for the given time period.\n\n    Args:\n        time_series (list):     The list of timeseries names to retrieve data for. Each timeseries can be either a string containing the\n                            ts name or a dictionary containing the ts name and a list of specific aggregate functions.\n\n        aggregates (list):  The list of aggregate functions you wish to apply to the data for which you have not\n                            specified an aggregate function. Valid aggregate functions are: 'average/avg, max, min,\n                            count, sum, interpolation/int, stepinterpolation/step'.\n\n        granularity (str):  The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,\n                            second/s', or a multiple of these indicated by a number as a prefix e.g. '12hour'.\n\n        start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.\n                                    E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since\n                                    epoch or a datetime object which will be converted to ms since epoch UTC.\n\n        end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.\n\n    Keyword Arguments:\n        api_key (str):                  Your api-key.\n\n        project (str):                  Project name.\n\n    Returns:\n        pandas.DataFrame: A pandas dataframe containing the datapoints for the given timeseries. The datapoints for all the\n        timeseries will all be on the same timestamps.\n\n    Note:\n        The ``timeseries`` parameter can take a list of strings and/or dicts on the following formats::\n\n            Using strings:\n                ['<timeseries1>', '<timeseries2>']\n\n            Using dicts:\n                [{'name': '<timeseries1>', 'aggregates': ['<aggfunc1>', '<aggfunc2>']},\n                {'name': '<timeseries2>', 'aggregates': []}]\n\n            Using both:\n                ['<timeseries1>', {'name': '<timeseries2>', 'aggregates': ['<aggfunc1>', '<aggfunc2>']}]\n    "
    (api_key, project) = (kwargs.get("api_key"), kwargs.get("project"))
    cookies = kwargs.get("cookies")
    url = config.get_base_url() + "/api/0.5/projects/{}/timeseries/dataframe".format(project)
    num_aggregates = 0
    for ts in time_series:
        if isinstance(ts, str) or (ts.get("aggregates") is None):
            num_aggregates += len(aggregates)
        else:
            num_aggregates += len(ts["aggregates"])
    per_tag_limit = int((_constants.LIMIT / num_aggregates))
    body = {
        "items": [
            (
                {"name": "{}".format(ts)}
                if isinstance(ts, str)
                else {"name": "{}".format(ts["name"]), "aggregates": ts.get("aggregates", [])}
            )
            for ts in time_series
        ],
        "aggregates": aggregates,
        "granularity": granularity,
        "start": start,
        "end": end,
        "limit": per_tag_limit,
    }
    headers = {"api-key": api_key, "content-type": "application/json", "accept": "text/csv"}
    dataframes = []
    while ((not dataframes) or (dataframes[(-1)].shape[0] == per_tag_limit)) and (body["end"] > body["start"]):
        res = _utils.post_request(url=url, body=body, headers=headers, cookies=cookies)
        dataframes.append(
            pd.read_csv(io.StringIO(res.content.decode((res.encoding if res.encoding else res.apparent_encoding))))
        )
        if dataframes[(-1)].empty:
            break
        latest_timestamp = int(dataframes[(-1)].iloc[((-1), 0)])
        body["start"] = latest_timestamp + _utils.granularity_to_ms(granularity)
    return pd.concat(dataframes).reset_index(drop=True)


def _get_datapoints_frame_user_defined_limit(time_series, aggregates, granularity, start, end, limit, **kwargs):
    "Returns a DatapointsResponse object with the requested data.\n\n    No paging or parallelizing is done.\n\n    Args:\n        time_series (str):       The list of timeseries names to retrieve data for. Each timeseries can be either a string containing the\n                            ts name or a dictionary containing the ts name and a list of specific aggregate functions.\n\n        aggregates (list):      The list of aggregate functions you wish to apply to the data. Valid aggregate functions\n                                are: 'average/avg, max, min, count, sum, interpolation/int, stepinterpolation/step'.\n\n        granularity (str):      The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,\n                                second/s', or a multiple of these indicated by a number as a prefix e.g. '12hour'.\n\n        start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.\n                                    E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since\n                                    epoch or a datetime object which will be converted to ms since epoch UTC.\n\n        end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.\n\n        limit (int):            Max number of rows to retrieve. Max is 100,000.\n\n    Keyword Arguments:\n        api_key (str):          Your api-key. Obligatory in this helper method.\n\n        project (str):          Project name. Obligatory in this helper method.\n    Returns:\n        v05.dto.DatapointsResponse: A data object containing the requested data with several getter methods with different\n        output formats.\n    "
    (api_key, project) = (kwargs.get("api_key"), kwargs.get("project"))
    cookies = kwargs.get("cookies")
    url = config.get_base_url() + "/api/0.5/projects/{}/timeseries/dataframe".format(project)
    body = {
        "items": [
            (
                {"name": "{}".format(ts)}
                if isinstance(ts, str)
                else {"name": "{}".format(ts["name"]), "aggregates": ts.get("aggregates", [])}
            )
            for ts in time_series
        ],
        "aggregates": aggregates,
        "granularity": granularity,
        "start": start,
        "end": end,
        "limit": limit,
    }
    headers = {"api-key": api_key, "content-type": "application/json", "accept": "text/csv"}
    res = _utils.post_request(url=url, body=body, headers=headers, cookies=cookies)
    df = pd.read_csv(io.StringIO(res.content.decode((res.encoding if res.encoding else res.apparent_encoding))))
    return df


def post_datapoints_frame(data, **kwargs):
    "Write a dataframe \n\n    Args:\n        dataframe (DataFrame):  Pandas DataFrame Object containing the timeseries\n\n    Keyword Args:\n        api_key (str): Your api-key.\n\n        project (str): Project name.\n\n    Returns:\n        An empty response.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    try:
        timestamp = data.timestamp
        names = data.drop(["timestamp"], axis=1).columns
    except:
        raise _utils.InputError("DataFrame not on a correct format")
    for name in names:
        data_points = [Datapoint(int(timestamp[i]), data[name].iloc[i]) for i in range(0, len(data))]
        res = post_datapoints(name, data_points, api_key=api_key, project=project)
    return res


def get_timeseries(prefix=None, description=None, include_metadata=False, asset_id=None, path=None, **kwargs):
    "Returns a TimeseriesObject containing the requested timeseries.\n\n    Args:\n        prefix (str):           List timeseries with this prefix in the name.\n\n        description (str):      Filter timeseries taht contains this string in its description.\n\n        include_metadata (bool):    Decide if the metadata field should be returned or not. Defaults to False.\n\n        asset_id (int):        Get timeseries related to this asset.\n\n        path (str):             Get timeseries under this asset path branch.\n\n    Keyword Arguments:\n        limit (int):            Number of results to return.\n\n        api_key (str):          Your api-key.\n\n        project (str):          Project name.\n\n        autopaging (bool):      Whether or not to automatically page through results. If set to true, limit will be\n                                disregarded. Defaults to False.\n\n    Returns:\n        v05.dto.TimeSeriesResponse: A data object containing the requested timeseries with several getter methods with different\n        output formats.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.5/projects/{}/timeseries".format(project)
    headers = {"api-key": api_key, "accept": "application/json"}
    params = {
        "q": prefix,
        "description": description,
        "includeMetadata": include_metadata,
        "assetId": asset_id,
        "path": path,
        "limit": (kwargs.get("limit", 10000) if (not kwargs.get("autopaging")) else 10000),
    }
    time_series = []
    res = _utils.get_request(url=url, headers=headers, params=params, cookies=config.get_cookies())
    time_series.extend(res.json()["data"]["items"])
    next_cursor = res.json()["data"].get("nextCursor")
    while next_cursor and kwargs.get("autopaging"):
        params["cursor"] = next_cursor
        res = _utils.get_request(url=url, headers=headers, params=params, cookies=config.get_cookies())
        time_series.extend(res.json()["data"]["items"])
        next_cursor = res.json()["data"].get("nextCursor")
    return TimeSeriesResponse(
        {
            "data": {
                "nextCursor": next_cursor,
                "previousCursor": res.json()["data"].get("previousCursor"),
                "items": time_series,
            }
        }
    )


def post_time_series(time_series, **kwargs):
    "Create a new time series.\n\n    Args:\n        time_series (list[v05.dto.TimeSeries]):   List of time series data transfer objects to create.\n\n    Keyword Args:\n        api_key (str): Your api-key.\n\n        project (str): Project name.\n    Returns:\n        An empty response.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.5/projects/{}/timeseries".format(project)
    body = {"items": [ts.__dict__ for ts in time_series]}
    headers = {"api-key": api_key, "content-type": "application/json", "accept": "application/json"}
    res = _utils.post_request(url, body=body, headers=headers)
    return res.json()


def update_time_series(time_series, **kwargs):
    "Update an existing time series.\n\n    For each field that can be updated, a null value indicates that nothing should be done.\n\n    Args:\n        time_series (list[v05.dto.TimeSeries]):   List of time series data transfer objects to update.\n\n    Keyword Args:\n        api_key (str): Your api-key.\n\n        project (str): Project name.\n\n    Returns:\n        An empty response.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.5/projects/{}/timeseries".format(project)
    body = {"items": [ts.__dict__ for ts in time_series]}
    headers = {"api-key": api_key, "content-type": "application/json", "accept": "application/json"}
    res = _utils.put_request(url, body=body, headers=headers)
    return res.json()


def delete_time_series(name, **kwargs):
    "Delete a timeseries.\n\n    Args:\n        name (str):   Name of timeseries to delete.\n\n    Keyword Args:\n        api_key (str): Your api-key.\n\n        project (str): Project name.\n\n    Returns:\n        An empty response.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.5/projects/{}/timeseries/{}".format(project, quote(name, safe=""))
    headers = {"api-key": api_key, "accept": "application/json"}
    res = _utils.delete_request(url, headers=headers)
    return res.json()


def live_data_generator(name, update_frequency=1, **kwargs):
    "Generator function which continously polls latest datapoint of a timeseries and yields new datapoints.\n\n    Args:\n        name (str): Name of timeseries to get latest datapoints for.\n\n        update_frequency (float): Frequency to pull for data in seconds.\n\n    Keyword Args:\n\n        api_key (str): Your api-key.\n\n        project (str): Project name.\n\n    Yields:\n        dict: Dictionary containing timestamp and value of latest datapoint.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    last_timestamp = get_latest(name, api_key=api_key, project=project).to_json()["timestamp"]
    while True:
        latest = get_latest(name, api_key=api_key, project=project).to_json()
        if last_timestamp == latest["timestamp"]:
            time.sleep(update_frequency)
        else:
            (yield latest)
        last_timestamp = latest["timestamp"]



"Events Module\n\nThis module mirrors the Events API. It allows you to get, post, update, and delete events.\n\nhttps://doc.cognitedata.com/0.5/#Cognite-API-Events\n"
import json
from cognite import _constants, _utils, config
from cognite.v05.dto import EventListResponse, EventResponse


def get_event(event_id, **kwargs):
    "Returns a EventResponse containing an event matching the id.\n\n    Args:\n        event_id (int):         The event id.\n\n    Keyword Arguments:\n        api_key (str):          Your api-key.\n\n        project (str):          Project name.\n\n    Returns:\n        v05.dto.EventResponse: A data object containing the requested event.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.5/projects/{}/events/{}".format(project, event_id)
    headers = {"api-key": api_key, "content-type": "application/json", "accept": "application/json"}
    res = _utils.get_request(url, headers=headers, cookies=config.get_cookies())
    return EventResponse(res.json())


def get_events(type=None, sub_type=None, asset_id=None, **kwargs):
    "Returns an EventListReponse object containing events matching the query.\n\n    Args:\n        type (str):             Type (class) of event, e.g. 'failure'.\n        sub_type (str):         Sub-type of event, e.g. 'electrical'.\n        asset_id (str):         Return events associated with this assetId.\n    Keyword Arguments:\n        sort (str):             Sort descending or ascending. Default 'ASC'.\n        cursor (str):           Cursor to use for paging through results.\n        limit (int):            Return up to this many results. Maximum is 10000. Default is 25.\n        has_description (bool): Return only events that have a textual description. Default null. False gives only\n                                those without description.\n        min_start_time (string): Only return events from after this time.\n        max_start_time (string): Only return events form before this time.\n        api_key (str):          Your api-key.\n        project (str):          Project name.\n        autopaging (bool):      Whether or not to automatically page through results. If set to true, limit will be\n                                disregarded. Defaults to False.\n\n    Returns:\n        v05.dto.EventListResponse: A data object containing the requested event.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.5/projects/{}/events".format(project)
    headers = {"api-key": api_key, "content-type": "application/json", "accept": "application/json"}
    if asset_id:
        params = {
            "assetId": asset_id,
            "sort": kwargs.get("sort"),
            "cursor": kwargs.get("cursor"),
            "limit": (kwargs.get("limit", 25) if (not kwargs.get("autopaging")) else _constants.LIMIT_AGG),
        }
    else:
        params = {
            "type": type,
            "subtype": sub_type,
            "assetId": asset_id,
            "sort": kwargs.get("sort"),
            "cursor": kwargs.get("cursor"),
            "limit": (kwargs.get("limit", 25) if (not kwargs.get("autopaging")) else _constants.LIMIT_AGG),
            "hasDescription": kwargs.get("has_description"),
            "minStartTime": kwargs.get("min_start_time"),
            "maxStartTime": kwargs.get("max_start_time"),
        }
    res = _utils.get_request(url, headers=headers, params=params, cookies=config.get_cookies())
    events = []
    events.extend(res.json()["data"]["items"])
    next_cursor = res.json()["data"].get("nextCursor")
    while next_cursor and kwargs.get("autopaging"):
        params["cursor"] = next_cursor
        res = _utils.get_request(url=url, headers=headers, params=params, cookies=config.get_cookies())
        events.extend(res.json()["data"]["items"])
        next_cursor = res.json()["data"].get("nextCursor")
    return EventListResponse(
        {
            "data": {
                "nextCursor": next_cursor,
                "previousCursor": res.json()["data"].get("previousCursor"),
                "items": events,
            }
        }
    )


def post_events(events, **kwargs):
    "Adds a list of events and returns an EventListResponse object containing created events.\n\n    Args:\n        events (List[v05.dto.Event]):    List of events to create.\n\n    Keyword Args:\n        api_key (str):          Your api-key.\n        project (str):          Project name.\n\n    Returns:\n        v05.dto.EventListResponse\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.5/projects/{}/events".format(project)
    headers = {"api-key": api_key, "content-type": "application/json", "accept": "application/json"}
    body = {"items": [event.__dict__ for event in events]}
    res = _utils.post_request(url, body=body, headers=headers)
    return EventListResponse(res.json())


def delete_events(event_ids, **kwargs):
    "Deletes a list of events.\n\n    Args:\n        event_ids (List[int]):    List of ids of events to delete.\n\n    Keyword Args:\n        api_key (str):          Your api-key.\n        project (str):          Project name.\n\n    Returns:\n        An empty response.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.5/projects/{}/events/delete".format(project)
    headers = {"api-key": api_key, "content-type": "application/json", "accept": "application/json"}
    body = {"items": event_ids}
    res = _utils.post_request(url, body=body, headers=headers)
    return res.json()


def search_for_events(
    description=None,
    type=None,
    subtype=None,
    min_start_time=None,
    max_start_time=None,
    min_end_time=None,
    max_end_time=None,
    min_created_time=None,
    max_created_time=None,
    min_last_updated_time=None,
    max_last_updated_time=None,
    metadata=None,
    asset_ids=None,
    asset_subtrees=None,
    **kwargs
):
    'Search for events.\n\n        Args:\n            description str:   Prefix and fuzzy search on description.\n\n        Keyword Args:\n            api_key (str):          Your api-key.\n            project (str):          Project name.\n            type (str):             Filter on type (case-sensitive).\n            subtype (str):          Filter on subtype (case-sensitive).\n            min_start_time (str):   Filter out events with startTime before this. Format is milliseconds since epoch.\n            max_start_time (str):   Filter out events with startTime after this. Format is milliseconds since epoch.\n            min_end_time (str):     Filter out events with endTime before this. Format is milliseconds since epoch.\n            max_end_time (str):     Filter out events with endTime after this. Format is milliseconds since epoch.\n            min_created_time(str):  Filter out events with createdTime before this. Format is milliseconds since epoch.\n            max_created_time (str): Filter out events with createdTime after this. Format is milliseconds since epoch.\n            min_last_updated_time(str):  Filter out events with lastUpdatedtime before this. Format is milliseconds since epoch.\n            max_last_updated_time(str): Filter out events with lastUpdatedtime after this. Format is milliseconds since epoch.\n            metadata (dict):        Filter out events that do not match these metadata fields and values (case-sensitive).\n                                    Format is {"key1":"value1","key2":"value2"}.\n            asset_ids (List[int]):  Filter out events that are not linked to any of these assets. Format is [12,345,6,7890].\n            asset_subtrees (List[int]): Filter out events that are not linked to assets in the subtree rooted at these assets.\n                                        Format is [12,345,6,7890].\n\n        Keyword Args:\n            sort (str):             Field to be sorted.\n            dir (str):              Sort direction (desc or asc)\n            limit (int):            Return up to this many results. Max is 1000, default is 25.\n            offset (int):           Offset from the first result. Sum of limit and offset must not exceed 1000. Default is 0.\n        Returns:\n            v05.dto.EventListResponse.\n        '
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.5/projects/{}/events/search".format(project)
    headers = {"api-key": api_key, "content-type": "application/json", "accept": "application/json"}
    params = {
        "description": description,
        "type": type,
        "subtype": subtype,
        "minStartTime": min_start_time,
        "maxStartTime": max_start_time,
        "minEndTime": min_end_time,
        "maxEndTime": max_end_time,
        "minCreatedTime": min_created_time,
        "maxCreatedTime": max_created_time,
        "minLastUpdatedTime": min_last_updated_time,
        "maxLastUpdatedTime": max_last_updated_time,
        "metadata": json.dumps(metadata),
        "assetIds": str((asset_ids or [])),
        "assetSubtrees": asset_subtrees,
        "sort": kwargs.get("sort"),
        "dir": kwargs.get("dir"),
        "limit": kwargs.get("limit", 1000),
        "offset": kwargs.get("offset"),
    }
    res = _utils.get_request(url, headers=headers, params=params)
    return EventListResponse(res.json())



import argparse
import importlib.util
import os
import shutil
import sys
from datetime import datetime
from distutils.core import run_setup

from tabulate import tabulate

from cognite.v06.analytics import models

MODEL_SRC_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir, "auxiliary/_hosting/model"))


class CogniteModelsCLI:
    def get(self, args):
        """cognite models get [<args>]

        Get all models or pass a model id and get all model versions.
        """
        parser = argparse.ArgumentParser(description="Get information about models.", usage=self.get.__doc__)
        parser.add_argument(
            "model_id", metavar="model-id", help="ID of model to get versions for", type=int, nargs="?", default=None
        )
        parser.add_argument("-s", "--source-packages", help="Get source-packages", action="store_true")
        parsed_args = parser.parse_args(args)
        if parsed_args.source_packages:
            source_package_list = models.get_source_packages()
            self._print_source_packages(source_package_list)
        elif parsed_args.model_id:
            versions_list = models.get_versions(parsed_args.model_id)
            self._print_versions(versions_list)
        else:
            models_list = models.get_models()
            self._print_models(models_list)

    def source(self, args):
        """cognite models source <source-package-name>

        Create a new model source package in the current working directory
        """
        parser = argparse.ArgumentParser(description="Create new model source package.", usage=self.source.__doc__)
        parser.add_argument("name", help="Name of the source package.")
        parsed_args = parser.parse_args(args)
        dest_dir = os.path.join(os.getcwd(), parsed_args.name)
        shutil.copytree(MODEL_SRC_DIR, dest_dir)
        print(
            "Your source package has been prepared!\nEnter the following command to take a look\n\n   cd {}\n".format(
                parsed_args.name
            )
        )

    def deploy(self, args):
        """cognite models deploy [<args>]

        1) Runs tests
        2) Builds sourcepackage
        3) Uploads sourcepackage
        4) Creates model (if --model-name is passed)
        """
        parser = argparse.ArgumentParser(description="Upload your source package", usage=self.deploy.__doc__)
        parser.add_argument("-m", "--model-name", help="Name of the model")
        parser.add_argument("-d", "--description", help="A description of what the model does", default="")
        parsed_args = parser.parse_args(args)

        print("* Verifying package...")
        info = self._verify_source_package(os.getcwd())

        print("* Building distribution...")
        run_setup(os.path.join(os.getcwd(), "setup.py"), script_args=["sdist"])

        print("* Uploading Source Package...")
        dist_dir = os.path.join(os.getcwd(), "dist")
        tarball_path = None
        for root, subdirs, files in os.walk(dist_dir):
            for f in files:
                if f.endswith(".tar.gz"):
                    tarball_path = os.path.join(dist_dir, f)

        source_package_id = models.create_source_package(
            info.get("source_package_name"),
            parsed_args.description,
            info.get("package_name"),
            info.get("available_operations"),
            file_path=tarball_path,
        )

        if parsed_args.model_name is not None:
            print("* Creating model...")
            model = models.create_model(name=parsed_args.model_name, description=parsed_args.description)

        print(
            "\nSuccesfully uploaded source package{}!".format(
                " and created {}".format(parsed_args.model_name) if parsed_args.model_name is not None else ""
            )
        )

        print("Source package ID: {}".format(source_package_id))
        if parsed_args.model_name is not None:
            print("Model ID: {}".format(model.get("id")))

    def _print_models(self, models_list):
        if not models_list:
            print("No models yet.")
            return

        models_tabular = [
            [
                m.get("id"),
                m.get("name"),
                m.get("description"),
                m.get("active_version_id"),
                datetime.fromtimestamp(int(m.get("created_time")) / 1000),
                m.get("is_deprecated"),
            ]
            for m in models_list
        ]
        headers = ["ID", "Name", "Description", "Active Version", "Created Time", "Is Deprecated"]
        print(tabulate(models_tabular, headers=headers))

    def _print_versions(self, versions_list):
        if not versions_list:
            print("No versions yet.")
            return
        versions_tabular = [
            [
                m.get("id"),
                m.get("name"),
                m.get("status"),
                m.get("description"),
                m.get("source_package_id"),
                datetime.fromtimestamp(int(m.get("created_time")) / 1000),
                m.get("is_deprecated"),
            ]
            for m in versions_list
        ]
        headers = ["ID", "Name", "Status", "Description", "Source Package ID", "Created Time", "Is Deprecated"]
        print(tabulate(versions_tabular, headers=headers))

    def _print_source_packages(self, source_package_list):
        if not source_package_list:
            print("No source packages yet.")
            return
        versions_tabular = [
            [
                m.get("id"),
                m.get("name"),
                m.get("description"),
                ", ".join(m.get("available_operations")),
                datetime.fromtimestamp(int(m.get("created_time")) / 1000),
                m.get("is_deprecated"),
                m.get("is_uploaded"),
            ]
            for m in source_package_list
        ]
        headers = ["ID", "Name", "Description", "Available Operations", "Created Time", "Is Deprecated", "Is Uploaded"]
        print(tabulate(versions_tabular, headers=headers))

    def _verify_source_package(self, directory):
        sys.path.append(os.getcwd())
        info = {"source_package_name": None, "package_name": None, "available_operations": []}

        has_model_module = False
        has_setup__file = False
        for root, subdirs, files in os.walk(directory):
            for f in files:
                if f == "model.py":
                    info["package_name"] = root.split("/")[-1]
                    has_model_module = True
                if f == "setup.py":
                    info["source_package_name"] = root.split("/")[-1]
                    has_setup__file = True

        if not has_setup__file:
            raise AssertionError("Your package must contain a setup.py file")
        if not has_model_module:
            raise AssertionError("Your package must contain a model.py file")

        spec = importlib.util.spec_from_file_location(
            "{}.model".format(info.get("package_name")), "./{}/model.py".format(info.get("package_name"))
        )
        model_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(model_module)

        if not hasattr(model_module, "Model"):
            raise AssertionError("model.py must contain a class named Model")

        model_class = model_module.Model

        has_train_routine = hasattr(model_class, "train")
        has_predict_routine = hasattr(model_class, "predict")
        has_load_routine = hasattr(model_class, "load")

        if not has_train_routine and not has_predict_routine:
            raise AssertionError("Your model class must define at least a train or predict routine.")

        if has_predict_routine and not has_load_routine:
            raise AssertionError("Your model class must define a 'load' method in order to perform predictions")

        if not has_train_routine:
            print(
                "* Your source package does not have a training routine and can not be used to train a new model version."
            )

        if not has_predict_routine:
            print(
                "* Your source package does not have a prediction routine and can not be used to make predictions on model versions."
            )
        if has_predict_routine:
            info.get("available_operations").append("predict")
        if has_train_routine:
            info.get("available_operations").append("train")

        return info



import re

from setuptools import find_packages, setup

packages = find_packages(exclude=["tests*"])

version = re.search('^__version__\s*=\s*"(.*)"', open("cognite/__init__.py").read(), re.M).group(1)

setup(
    name="cognite-sdk",
    version=version,
    # entry_points={"console_scripts": ["cognite = cognite.cli:main"]}, # TODO: enable when CLI is ready
    description="Cognite API SDK for Python",
    url="http://cognite-sdk-python.readthedocs.io/",
    download_url="https://github.com/cognitedata/cognite-sdk-python/archive/{}.tar.gz".format(version),
    author="Erlend Vollset",
    author_email="erlend.vollset@cognite.com",
    packages=packages,
    install_requires=["requests", "pandas", "protobuf", "cognite-logger>=0.3", "tabulate"],
    python_requires=">=3.3",
    zip_safe=False,
    include_package_data=True,
)



"Raw Module\n\nThis module mirrors the Raw API. It allows the user to handle raw data.\n\nhttps://doc.cognitedata.com/0.4/#Cognite-API-Cloud-Raw\n"
import cognite._utils as _utils
import cognite.config as config
from cognite.v04.dto import RawResponse, RawRow


def get_databases(limit=None, cursor=None, api_key=None, project=None):
    "Returns a RawObject containing a list of raw databases.\n\n    Args:\n        limit (int):    A limit on the amount of results to return.\n\n        cursor (str):   A cursor can be provided to navigate through pages of results.\n\n        api_key (str):  Your api-key.\n\n        project (str):  Project name.\n\n    Returns:\n        v04.dto.RawResponse: A data object containing the requested data with several getter methods with different\n        output formats.\n    "
    (api_key, project) = config.get_config_variables(api_key, project)
    url = config.get_base_url() + "/api/0.4/projects/{}/raw".format(project)
    params = {"limit": limit, "cursor": cursor}
    headers = {"api-key": api_key, "content-type": "*/*", "accept": "application/json"}
    res = _utils.get_request(url=url, params=params, headers=headers, cookies=config.get_cookies())
    return RawResponse(res.json())


def create_databases(database_names, api_key=None, project=None):
    "Creates databases in the Raw API and returns the created databases.\n\n    Args:\n        database_names (list):  A list of databases to create.\n\n        api_key (str):          Your api-key.\n\n        project (str):          Project name.\n\n    Returns:\n        v04.dto.RawResponse: A data object containing the requested data with several getter methods with different\n        output formats.\n\n    "
    (api_key, project) = config.get_config_variables(api_key, project)
    url = config.get_base_url() + "/api/0.4/projects/{}/raw/create".format(project)
    body = {"items": [{"dbName": "{}".format(database_name)} for database_name in database_names]}
    headers = {"api-key": api_key, "content-type": "*/*", "accept": "application/json"}
    res = _utils.post_request(url=url, body=body, headers=headers, cookies=config.get_cookies())
    return RawResponse(res.json())


def delete_databases(database_names, recursive=False, api_key=None, project=None):
    "Deletes databases in the Raw API.\n\n    Args:\n        database_names (list):  A list of databases to delete.\n\n        api_key (str):          Your api-key.\n\n        project (str):          Project name.\n\n    Returns:\n        An empty response.\n\n    "
    (api_key, project) = config.get_config_variables(api_key, project)
    url = config.get_base_url() + "/api/0.4/projects/{}/raw/delete".format(project)
    body = {"items": [{"dbName": "{}".format(database_name)} for database_name in database_names]}
    params = {"recursive": recursive}
    headers = {"api-key": api_key, "content-type": "*/*", "accept": "application/json"}
    res = _utils.post_request(url=url, body=body, params=params, headers=headers, cookies=config.get_cookies())
    return res.json()


def get_tables(database_name=None, limit=None, cursor=None, api_key=None, project=None):
    "Returns a RawObject containing a list of tables in a raw database.\n\n    Args:\n        database_name (str):   The database name to retrieve tables from.\n\n        limit (int):    A limit on the amount of results to return.\n\n        cursor (str):   A cursor can be provided to navigate through pages of results.\n\n        api_key (str):  Your api-key.\n\n        project (str):  Project name.\n\n    Returns:\n        v04.dto.RawResponse: A data object containing the requested data with several getter methods with different\n        output formats.\n    "
    (api_key, project) = config.get_config_variables(api_key, project)
    url = config.get_base_url() + "/api/0.4/projects/{}/raw/{}".format(project, database_name)
    params = dict()
    if not limit:
        params["limit"] = limit
    if not cursor:
        params["cursor"] = cursor
    headers = {"api-key": api_key, "content-type": "*/*", "accept": "application/json"}
    res = _utils.get_request(url=url, params=params, headers=headers, cookies=config.get_cookies())
    return RawResponse(res.json())


def create_tables(database_name=None, table_names=None, api_key=None, project=None):
    "Creates tables in the given Raw API database.\n\n    Args:\n        database_name (str):    The database to create tables in.\n\n        table_names (list):     The table names to create.\n\n        api_key (str):          Your api-key.\n\n        project (str):          Project name.\n\n    Returns:\n        v04.dto.RawResponse: A data object containing the requested data with several getter methods with different\n        output formats.\n\n    "
    (api_key, project) = config.get_config_variables(api_key, project)
    url = config.get_base_url() + "/api/0.4/projects/{}/raw/{}/create".format(project, database_name)
    body = {"items": [{"tableName": "{}".format(table_name)} for table_name in table_names]}
    headers = {"api-key": api_key, "content-type": "*/*", "accept": "application/json"}
    res = _utils.post_request(url=url, body=body, headers=headers, cookies=config.get_cookies())
    return RawResponse(res.json())


def delete_tables(database_name=None, table_names=None, api_key=None, project=None):
    "Deletes databases in the Raw API.\n\n    Args:\n        database_name (str):    The database to create tables in.\n\n        table_names (list):     The table names to create.\n\n        api_key (str):          Your api-key.\n\n        project (str):          Project name.\n\n    Returns:\n        An empty response.\n\n    "
    (api_key, project) = config.get_config_variables(api_key, project)
    url = config.get_base_url() + "/api/0.4/projects/{}/raw/{}/delete".format(project, database_name)
    body = {"items": [{"tableName": "{}".format(table_name)} for table_name in table_names]}
    headers = {"api-key": api_key, "content-type": "*/*", "accept": "application/json"}
    res = _utils.post_request(url=url, body=body, headers=headers, cookies=config.get_cookies())
    return res.json()


def get_rows(database_name=None, table_name=None, limit=None, cursor=None, api_key=None, project=None):
    "Returns a RawResponse Object containing a list of rows.\n\n    Args:\n        database_name (str):    The database name to retrieve rows from.\n\n        table_name (str):       The table name to retrieve rows from.\n\n        limit (int):            A limit on the amount of results to return.\n\n        cursor (str):           A cursor can be provided to navigate through pages of results.\n\n        api_key (str):          Your api-key.\n\n        project (str):          Project name.\n\n    Returns:\n        v04.dto.RawResponse: A data object containing the requested data with several getter methods with different\n        output formats.\n    "
    (api_key, project) = config.get_config_variables(api_key, project)
    url = config.get_base_url() + "/api/0.4/projects/{}/raw/{}/{}".format(project, database_name, table_name)
    params = dict()
    params["limit"] = limit
    params["cursor"] = cursor
    headers = {"api-key": api_key, "content-type": "*/*", "accept": "application/json"}
    res = _utils.get_request(url=url, params=params, headers=headers, cookies=config.get_cookies())
    return RawResponse(res.json())


def create_rows(
    database_name=None, table_name=None, rows=None, api_key=None, project=None, ensure_parent=False, use_gzip=False
):
    "Creates tables in the given Raw API database.\n\n    Args:\n        database_name (str):    The database to create rows in.\n\n        table_name (str):       The table names to create rows in.\n\n        rows (list[v04.dto.RawRow]):            The rows to create.\n\n        api_key (str):          Your api-key.\n\n        project (str):          Project name.\n\n        ensure_parent (bool):   Create database/table if it doesn't exist already\n\n        use_gzip (bool):        Compress content using gzip\n\n    Returns:\n        An empty response\n\n    "
    (api_key, project) = config.get_config_variables(api_key, project)
    url = config.get_base_url() + "/api/0.4/projects/{}/raw/{}/{}/create".format(project, database_name, table_name)
    headers = {"api-key": api_key, "content-type": "*/*", "accept": "application/json"}
    if ensure_parent:
        params = {"ensureParent": "true"}
    else:
        params = {}
    ul_row_limit = 1000
    i = 0
    while i < len(rows):
        body = {
            "items": [{"key": "{}".format(row.key), "columns": row.columns} for row in rows[i : (i + ul_row_limit)]]
        }
        res = _utils.post_request(
            url=url, body=body, headers=headers, params=params, cookies=config.get_cookies(), use_gzip=use_gzip
        )
        i += ul_row_limit
    return res.json()


def delete_rows(database_name=None, table_name=None, rows=None, api_key=None, project=None):
    "Deletes rows in the Raw API.\n\n    Args:\n        database_name (str):    The database to create tables in.\n\n        table_name (str):      The table name where the rows are at.\n\n        rows (list):            The rows to delete.\n\n        api_key (str):          Your api-key.\n\n        project (str):          Project name.\n\n    Returns:\n        An empty response.\n\n    "
    (api_key, project) = config.get_config_variables(api_key, project)
    url = config.get_base_url() + "/api/0.4/projects/{}/raw/{}/{}/delete".format(project, database_name, table_name)
    body = {"items": [{"key": "{}".format(row.key), "columns": row.columns} for row in rows]}
    headers = {"api-key": api_key, "content-type": "*/*", "accept": "application/json"}
    res = _utils.post_request(url=url, body=body, headers=headers, cookies=config.get_cookies())
    return res.json()


def get_row(database_name=None, table_name=None, row_key=None, api_key=None, project=None):
    "Returns a RawResponse Object containing a list of rows.\n\n    Args:\n        database_name (str):    The database name to retrieve rows from.\n\n        table_name (str):       The table name to retrieve rows from.\n\n        row_key (str):          The key of the row to fetch.\n\n        api_key (str):          Your api-key.\n\n        project (str):          Project name.\n\n    Returns:\n        v04.dto.RawResponse: A data object containing the requested data with several getter methods with different\n        output formats.\n    "
    (api_key, project) = config.get_config_variables(api_key, project)
    url = config.get_base_url() + "/api/0.4/projects/{}/raw/{}/{}/{}".format(
        project, database_name, table_name, row_key
    )
    params = dict()
    headers = {"api-key": api_key, "content-type": "*/*", "accept": "application/json"}
    res = _utils.get_request(url=url, params=params, headers=headers, cookies=config.get_cookies())
    return RawResponse(res.json())



"Cognite API SDK\n\nThis package has been created to ensure excellent excellent user experience for data scientists using the Cognite\nData Platform (CDP).\n"
__all__ = ["v04", "v05", "v06", "preprocessing", "config", "data_transfer_service"]
__version__ = "0.11.15"



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api_timeseries_data_v1.proto

import sys

_b = sys.version_info[0] < 3 and (lambda x: x) or (lambda x: x.encode("latin1"))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="api_timeseries_data_v1.proto",
    package="api",
    syntax="proto3",
    serialized_pb=_b(
        '\n\x1c\x61pi_timeseries_data_v1.proto\x12\x03\x61pi"3\n\x0fStringDatapoint\x12\x11\n\ttimestamp\x18\x01 \x01(\x03\x12\r\n\x05value\x18\x02 \x01(\t"4\n\x10NumericDatapoint\x12\x11\n\ttimestamp\x18\x01 \x01(\x03\x12\r\n\x05value\x18\x02 \x01(\x01"<\n\x14StringTimeseriesData\x12$\n\x06points\x18\x01 \x03(\x0b\x32\x14.api.StringDatapoint">\n\x15NumericTimeseriesData\x12%\n\x06points\x18\x01 \x03(\x0b\x32\x15.api.NumericDatapoint"|\n\x0eTimeseriesData\x12/\n\nstringData\x18\x01 \x01(\x0b\x32\x19.api.StringTimeseriesDataH\x00\x12\x31\n\x0bnumericData\x18\x02 \x01(\x0b\x32\x1a.api.NumericTimeseriesDataH\x00\x42\x06\n\x04\x64\x61ta"\x8e\x01\n\x11TagTimeseriesData\x12\r\n\x05tagId\x18\x01 \x01(\t\x12/\n\nstringData\x18\x02 \x01(\x0b\x32\x19.api.StringTimeseriesDataH\x00\x12\x31\n\x0bnumericData\x18\x03 \x01(\x0b\x32\x1a.api.NumericTimeseriesDataH\x00\x42\x06\n\x04\x64\x61ta"K\n\x16MultiTagTimeseriesData\x12\x31\n\x11tagTimeseriesData\x18\x01 \x03(\x0b\x32\x16.api.TagTimeseriesDataB4\n\x17\x63om.cognite.data.api.v1P\x01\xb8\x01\x01\xaa\x02\x13\x43ognite.Data.Api.V1b\x06proto3'
    ),
)


_STRINGDATAPOINT = _descriptor.Descriptor(
    name="StringDatapoint",
    full_name="api.StringDatapoint",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="timestamp",
            full_name="api.StringDatapoint.timestamp",
            index=0,
            number=1,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="value",
            full_name="api.StringDatapoint.value",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=37,
    serialized_end=88,
)


_NUMERICDATAPOINT = _descriptor.Descriptor(
    name="NumericDatapoint",
    full_name="api.NumericDatapoint",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="timestamp",
            full_name="api.NumericDatapoint.timestamp",
            index=0,
            number=1,
            type=3,
            cpp_type=2,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="value",
            full_name="api.NumericDatapoint.value",
            index=1,
            number=2,
            type=1,
            cpp_type=5,
            label=1,
            has_default_value=False,
            default_value=float(0),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=90,
    serialized_end=142,
)


_STRINGTIMESERIESDATA = _descriptor.Descriptor(
    name="StringTimeseriesData",
    full_name="api.StringTimeseriesData",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="points",
            full_name="api.StringTimeseriesData.points",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        )
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=144,
    serialized_end=204,
)


_NUMERICTIMESERIESDATA = _descriptor.Descriptor(
    name="NumericTimeseriesData",
    full_name="api.NumericTimeseriesData",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="points",
            full_name="api.NumericTimeseriesData.points",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        )
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=206,
    serialized_end=268,
)


_TIMESERIESDATA = _descriptor.Descriptor(
    name="TimeseriesData",
    full_name="api.TimeseriesData",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="stringData",
            full_name="api.TimeseriesData.stringData",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="numericData",
            full_name="api.TimeseriesData.numericData",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="data", full_name="api.TimeseriesData.data", index=0, containing_type=None, fields=[]
        )
    ],
    serialized_start=270,
    serialized_end=394,
)


_TAGTIMESERIESDATA = _descriptor.Descriptor(
    name="TagTimeseriesData",
    full_name="api.TagTimeseriesData",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="tagId",
            full_name="api.TagTimeseriesData.tagId",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="stringData",
            full_name="api.TagTimeseriesData.stringData",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="numericData",
            full_name="api.TagTimeseriesData.numericData",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="data", full_name="api.TagTimeseriesData.data", index=0, containing_type=None, fields=[]
        )
    ],
    serialized_start=397,
    serialized_end=539,
)


_MULTITAGTIMESERIESDATA = _descriptor.Descriptor(
    name="MultiTagTimeseriesData",
    full_name="api.MultiTagTimeseriesData",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="tagTimeseriesData",
            full_name="api.MultiTagTimeseriesData.tagTimeseriesData",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
            file=DESCRIPTOR,
        )
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=541,
    serialized_end=616,
)

_STRINGTIMESERIESDATA.fields_by_name["points"].message_type = _STRINGDATAPOINT
_NUMERICTIMESERIESDATA.fields_by_name["points"].message_type = _NUMERICDATAPOINT
_TIMESERIESDATA.fields_by_name["stringData"].message_type = _STRINGTIMESERIESDATA
_TIMESERIESDATA.fields_by_name["numericData"].message_type = _NUMERICTIMESERIESDATA
_TIMESERIESDATA.oneofs_by_name["data"].fields.append(_TIMESERIESDATA.fields_by_name["stringData"])
_TIMESERIESDATA.fields_by_name["stringData"].containing_oneof = _TIMESERIESDATA.oneofs_by_name["data"]
_TIMESERIESDATA.oneofs_by_name["data"].fields.append(_TIMESERIESDATA.fields_by_name["numericData"])
_TIMESERIESDATA.fields_by_name["numericData"].containing_oneof = _TIMESERIESDATA.oneofs_by_name["data"]
_TAGTIMESERIESDATA.fields_by_name["stringData"].message_type = _STRINGTIMESERIESDATA
_TAGTIMESERIESDATA.fields_by_name["numericData"].message_type = _NUMERICTIMESERIESDATA
_TAGTIMESERIESDATA.oneofs_by_name["data"].fields.append(_TAGTIMESERIESDATA.fields_by_name["stringData"])
_TAGTIMESERIESDATA.fields_by_name["stringData"].containing_oneof = _TAGTIMESERIESDATA.oneofs_by_name["data"]
_TAGTIMESERIESDATA.oneofs_by_name["data"].fields.append(_TAGTIMESERIESDATA.fields_by_name["numericData"])
_TAGTIMESERIESDATA.fields_by_name["numericData"].containing_oneof = _TAGTIMESERIESDATA.oneofs_by_name["data"]
_MULTITAGTIMESERIESDATA.fields_by_name["tagTimeseriesData"].message_type = _TAGTIMESERIESDATA
DESCRIPTOR.message_types_by_name["StringDatapoint"] = _STRINGDATAPOINT
DESCRIPTOR.message_types_by_name["NumericDatapoint"] = _NUMERICDATAPOINT
DESCRIPTOR.message_types_by_name["StringTimeseriesData"] = _STRINGTIMESERIESDATA
DESCRIPTOR.message_types_by_name["NumericTimeseriesData"] = _NUMERICTIMESERIESDATA
DESCRIPTOR.message_types_by_name["TimeseriesData"] = _TIMESERIESDATA
DESCRIPTOR.message_types_by_name["TagTimeseriesData"] = _TAGTIMESERIESDATA
DESCRIPTOR.message_types_by_name["MultiTagTimeseriesData"] = _MULTITAGTIMESERIESDATA
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

StringDatapoint = _reflection.GeneratedProtocolMessageType(
    "StringDatapoint",
    (_message.Message,),
    dict(
        DESCRIPTOR=_STRINGDATAPOINT,
        __module__="api_timeseries_data_v1_pb2"
        # @@protoc_insertion_point(class_scope:api.StringDatapoint)
    ),
)
_sym_db.RegisterMessage(StringDatapoint)

NumericDatapoint = _reflection.GeneratedProtocolMessageType(
    "NumericDatapoint",
    (_message.Message,),
    dict(
        DESCRIPTOR=_NUMERICDATAPOINT,
        __module__="api_timeseries_data_v1_pb2"
        # @@protoc_insertion_point(class_scope:api.NumericDatapoint)
    ),
)
_sym_db.RegisterMessage(NumericDatapoint)

StringTimeseriesData = _reflection.GeneratedProtocolMessageType(
    "StringTimeseriesData",
    (_message.Message,),
    dict(
        DESCRIPTOR=_STRINGTIMESERIESDATA,
        __module__="api_timeseries_data_v1_pb2"
        # @@protoc_insertion_point(class_scope:api.StringTimeseriesData)
    ),
)
_sym_db.RegisterMessage(StringTimeseriesData)

NumericTimeseriesData = _reflection.GeneratedProtocolMessageType(
    "NumericTimeseriesData",
    (_message.Message,),
    dict(
        DESCRIPTOR=_NUMERICTIMESERIESDATA,
        __module__="api_timeseries_data_v1_pb2"
        # @@protoc_insertion_point(class_scope:api.NumericTimeseriesData)
    ),
)
_sym_db.RegisterMessage(NumericTimeseriesData)

TimeseriesData = _reflection.GeneratedProtocolMessageType(
    "TimeseriesData",
    (_message.Message,),
    dict(
        DESCRIPTOR=_TIMESERIESDATA,
        __module__="api_timeseries_data_v1_pb2"
        # @@protoc_insertion_point(class_scope:api.TimeseriesData)
    ),
)
_sym_db.RegisterMessage(TimeseriesData)

TagTimeseriesData = _reflection.GeneratedProtocolMessageType(
    "TagTimeseriesData",
    (_message.Message,),
    dict(
        DESCRIPTOR=_TAGTIMESERIESDATA,
        __module__="api_timeseries_data_v1_pb2"
        # @@protoc_insertion_point(class_scope:api.TagTimeseriesData)
    ),
)
_sym_db.RegisterMessage(TagTimeseriesData)

MultiTagTimeseriesData = _reflection.GeneratedProtocolMessageType(
    "MultiTagTimeseriesData",
    (_message.Message,),
    dict(
        DESCRIPTOR=_MULTITAGTIMESERIESDATA,
        __module__="api_timeseries_data_v1_pb2"
        # @@protoc_insertion_point(class_scope:api.MultiTagTimeseriesData)
    ),
)
_sym_db.RegisterMessage(MultiTagTimeseriesData)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(
    descriptor_pb2.FileOptions(), _b("\n\027com.cognite.data.api.v1P\001\270\001\001\252\002\023Cognite.Data.Api.V1")
)
# @@protoc_insertion_point(module_scope)



"Assets Module.\n\nThis module mirrors the Assets API.\n\nhttps://doc.cognitedata.com/0.4/#Cognite-API-Assets\n"
import cognite._constants as constants
import cognite._utils as utils
import cognite.config as config
from cognite.v04.dto import AssetResponse, Asset


def get_assets(name=None, path=None, description=None, metadata=None, depth=None, fuzziness=None, **kwargs):
    "Returns assets matching provided description.\n\n    Args:\n        name (str):             The name of the asset(s) to get.\n\n        path (str):             The path of the subtree to search in.\n\n        description (str):      Search query.\n\n        metadata (str):         The metadata values used to filter the results.\n\n        depth (int):            Get sub assets up oto this many levels below the specified path.\n\n        fuzziness (int):        The degree of fuzziness in the name matching.\n\n    Keyword Arguments:\n        limit (int):            The maximum number of assets to be returned.\n\n        cursor (str):           Cursor to use for paging through results.\n\n        api_key (str):          Your api-key.\n\n        project (str):          Project name.\n    Returns:\n        v04.dto.AssetResponse: A data object containing the requested assets with several getter methods with different\n        output formats.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.4/projects/{}/assets".format(project)
    params = {
        "name": name,
        "description": description,
        "path": path,
        "metadata": metadata,
        "depth": depth,
        "fuzziness": fuzziness,
        "cursor": kwargs.get("cursor"),
        "limit": kwargs.get("limit", constants.LIMIT),
    }
    headers = {"api-key": api_key, "accept": "application/json"}
    res = utils.get_request(url, params=params, headers=headers, cookies=config.get_cookies())
    return AssetResponse(res.json())


def get_asset_subtree(asset_id="", depth=None, **kwargs):
    "Returns assets with provided assetId.\n\n    Args:\n        asset_id (str):         The asset id of the top asset to get.\n\n        depth (int):            Get subassets this many levels below the top asset.\n\n    Keyword Arguments:\n        limit (int):            The maximum nuber of assets to be returned.\n\n        cursor (str):           Cursor to use for paging through results.\n\n        api_key (str):          Your api-key.\n\n        project (str):          Project name.\n    Returns:\n        v04.dto.AssetResponse: A data object containing the requested assets with several getter methods with different\n        output formats.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.4/projects/{}/assets/{}".format(project, asset_id)
    params = {"depth": depth, "limit": kwargs.get("limit", constants.LIMIT), "cursor": kwargs.get("cursor")}
    headers = {"api-key": api_key, "accept": "application/json"}
    res = utils.get_request(url, params=params, headers=headers, cookies=config.get_cookies())
    return AssetResponse(res.json())


def post_assets(assets, **kwargs):
    "Insert a list of assets.\n\n    Args:\n        assets (list[v04.dto.Asset]): List of asset data transfer objects.\n\n    Keyword Args:\n        api_key (str): Your api-key.\n\n        project (str): Project name.\n\n    Returns:\n        v04.dto.AssetResponse: A data object containing the posted assets with several getter methods with different\n        output formats.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.4/projects/{}/assets".format(project)
    body = {"items": [asset.__dict__ for asset in assets]}
    headers = {"api-key": api_key, "content-type": "application/json", "accept": "application/json"}
    res = utils.post_request(url, body=body, headers=headers, cookies=config.get_cookies())
    return AssetResponse(res.json())


def delete_assets(asset_ids, **kwargs):
    "Delete a list of assets.\n\n    Args:\n        asset_ids (list[v04.dto.Asset]): List of IDs of assets to delete.\n\n    Keyword Args:\n        api_key (str): Your api-key.\n\n        project (str): Project name.\n\n    Returns:\n        An empty response.\n    "
    (api_key, project) = config.get_config_variables(kwargs.get("api_key"), kwargs.get("project"))
    url = config.get_base_url() + "/api/0.4/projects/{}/assets/delete".format(project)
    body = {"items": asset_ids}
    headers = {"api-key": api_key, "content-type": "application/json", "accept": "application/json"}
    res = utils.post_request(url, body=body, headers=headers, cookies=config.get_cookies())
    return res.json()
